﻿using System.Globalization;
using System.IdentityModel.Tokens.Jwt;
using System.Security.Claims;
using Infrastructure.Utils;
using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.IdentityModel.Tokens;

namespace Infrastructure.Security;

public interface ITokenBuilder
{
    IList<Claim> GetClaimsByUserContext(IUserContext userContext);

    void SetUserContext(TokenValidatedContext context);

    TokenInfo GenerateJwtTokenInfo(IReadOnlyCollection<Claim> claims);
}

public class TokenBuilder(
    JwtOptions jwtOptions,
    JwtSecurityTokenHandler jwtSecurityTokenHandler,
    IEncryptionService encryptionService)
    : ITokenBuilder
{
    public IList<Claim> GetClaimsByUserContext(IUserContext userContext)
    {
        var claims = new List<Claim>()
        {
            new(JwtRegisteredClaimNames.UniqueName, userContext.Username),
            new(JwtRegisteredClaimNames.NameId, userContext.Id.ToString()),
            new(JwtRegisteredClaimNames.Name, userContext.Name),
            new(JwtRegisteredClaimNames.Email, userContext.Email),
            new(JwtRegisteredClaimNames.Iat,
                EpochTime.GetIntDate(DateTime.Now).ToString(CultureInfo.InvariantCulture),
                ClaimValueTypes.Integer64),
            new(JwtRegisteredClaimNames.Exp, jwtOptions.Expiration.ToString())
        };
        claims.AddRange(userContext.RoleIds.Select(rId => new Claim(ClaimTypes.Role, rId)));
        return claims;
    }

    public void SetUserContext(TokenValidatedContext context)
    {
        var userContext = context.HttpContext.RequestServices.GetService<IUserContext>() ??
                          throw new NullReferenceException(nameof(IUserContext));
        var principal = context.Principal ?? throw new NullReferenceException(nameof(context.Principal));
        userContext.Id = long.Parse(
            principal.Claims.FirstOrDefault(c => c.Type == JwtRegisteredClaimNames.NameId)?.Value);
        userContext.Username =
            principal.Claims.FirstOrDefault(c => c.Type == JwtRegisteredClaimNames.UniqueName)?.Value!;
        userContext.Name = principal.Claims.FirstOrDefault(c => c.Type == JwtRegisteredClaimNames.Name)?.Value!;
        userContext.Email = principal.Claims.FirstOrDefault(c => c.Type == JwtRegisteredClaimNames.Email)?.Value!;
        userContext.RoleIds = principal.Claims.Where(c => c.Type == ClaimTypes.Role).Select(c => c.Value).ToArray();
        userContext.RemoteIpAddress = context.HttpContext.GetRequestIp();
    }

    public TokenInfo GenerateJwtTokenInfo(IReadOnlyCollection<Claim> claims)
    {
        var jwtToken = new JwtSecurityToken(
            issuer: jwtOptions.Issuer,
            audience: jwtOptions.Audience,
            claims: claims,
            notBefore: DateTime.Now,
            expires: DateTime.Now.AddSeconds(jwtOptions.Expiration),
            signingCredentials: jwtOptions.SigningCredentials);
        var token = jwtSecurityTokenHandler.WriteToken(jwtToken);
        return new TokenInfo(encryptionService.Encrypt(token), jwtOptions.Expiration,
            JwtBearerDefaults.AuthenticationScheme);
    }
}